/*
LibINI - INI-format I/O library for the CLR.
Copyright  2006-2007 Serenity Project  

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
using System;
using System.Collections.Generic;
using System.Text;

namespace LibINI
{
    /// <summary>
    /// Represents a single name/value pair inside of an IniSection. This class cannot be inherited.
    /// </summary>
    public sealed class IniEntry
    {
        #region Constructors - Internal
        internal IniEntry(string name, IniFile file)
        {
            this.name = name;
            this.file = file;
        }
        #endregion
        #region Fields - Private
        private string value;
        private string comment;
        private string name;
        private IniFile file;
        #endregion
        #region Methods - Public
        /// <summary>
        /// Creates and returns a deep clone of the current IniEntry.
        /// </summary>
        /// <returns></returns>
        public IniEntry Clone()
        {
            return this.Clone(this.File);
        }
        /// <summary>
        /// Creates and returns a deep clone of the current IniEntry.
        /// </summary>
        /// <returns>A deep clone of the current IniEntry.</returns>
        public IniEntry Clone(IniFile file)
        {
            IniEntry Clone = new IniEntry(this.Name, file);
            Clone.value = this.value;
            Clone.comment = this.comment;

            return Clone;
        }
        /// <summary>
        /// Creates a string representation of the current IniEntry.
        /// </summary>
        /// <returns>A string representation of the current IniEntry.</returns>
        public override string ToString()
        {
            StringBuilder output = new StringBuilder();
            if (this.file.WriteComments == true)
            {
                output.Append(this.comment + Environment.NewLine);
            }
            output.Append(this.Name + "=" + this.value + Environment.NewLine);

            return output.ToString();
        }
        #endregion
        #region Properties - Public
        /// <summary>
        /// Gets or sets a comment associated with the current IniEntry.
        /// </summary>
        public string Comment
        {
            get
            {
                return this.comment;
            }
            set
            {
                this.comment = value;
            }
        }
        /// <summary>
        /// Gets the IniFile that owns the current IniEntry.
        /// </summary>
        public IniFile File
        {
            get
            {
                return this.File;
            }
        }
        /// <summary>
        /// Gets the name of the current IniEntry.
        /// </summary>
        public string Name
        {
            get
            {
                return this.name;
            }
            internal set
            {
                if (IniFile.IsValidName(value) == true)
                {
                    this.name = value;
                }
                else
                {
                    throw new IniOperationException("An attempt was made to assign an invalid name to an IniEntry.", value);
                }
            }
        }
        /// <summary>
        /// Gets or sets the value portion of the current IniEntry.
        /// </summary>
        public string Value
        {
            get
            {
                return this.value;
            }
            set
            {
                if (IniFile.IsValidValue(value) == false)
                {
                    throw new IniOperationException("Value string contains invalid characters.", value);
                }
                else
                {
                    this.value = value;
                }
            }
        }
        #endregion
    }
}
