/*
LibINI - INI-format I/O library for the CLR.
Copyright  2006-2007 Serenity Project  

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace LibINI
{
    /// <summary>
    /// Represents an INI-format configuration settings file, and contains methods to add,
    /// remove, and modify configuration settings. This class cannot be inherited.
    /// </summary>
    /// <remarks>
    /// This class provides the core functionality of LibINI.
    /// </remarks>
    public sealed class IniFile : IEnumerable<IniSection>
    {
        #region Constructors - Public
        /// <summary>
        /// Initializes a new instance of the File class.
        /// </summary>
        /// <param name="location">The location where the new File object will
        /// read from or written to when Read() and Write() methods are invoked.</param>
        public IniFile(string location)
        {
            this.location = location;
            this.sections = new Dictionary<string, IniSection>();
            this.unsectioned = new IniSection("NoSection", this);
        }
        #endregion
        #region Fields - Private
        private bool caseSensitiveRetrieval = false;
        private char commentChar = ';';
        private IniSection lastSection;
        private Dictionary<string, IniSection> sections;
        private string location;
        private bool outputSorting = false;
        private IniSection unsectioned;
        private bool writeComments = true;
        private bool writeEmptySections = true;
        private Encoding readEncoding;
        private Encoding writeEncoding;
        #endregion
        #region Fields - Public
        /// <summary>
        /// Gets the chars that cannot be used for an IniSection or IniEntry name.
        /// </summary>
        public static readonly char[] InvalidNameChars = new char[] { '\r', '\n', '\t', '[', ';', '#' };
        /// <summary>
        /// Gets a constant string containing chars that cannot be used for a IniEntry value.
        /// </summary>
        public static readonly char[] InvalidValueChars = new char[] { '\r', '\n', '=', ';', '#' };
        #endregion
        #region Indexers - Public
        /// <summary>
        /// Gets or sets an IniSection in the current File.
        /// When setting an IniSection that does not exist in the current File, it will be added.
        /// When getting an IniSection that does not exist in the current File, a null will be returned.
        /// </summary>
        /// <param name="name">The name of the IniSection to get or set.</param>
        public IniSection this[string name]
        {
            get
            {
                if ((this.lastSection != null) && (name == this.lastSection.Name))
                {
                    return this.lastSection;
                }
                else if (this.sections.ContainsKey(name) == true)
                {
                    this.lastSection = this.sections[name];
                    return this.lastSection;
                }
                else
                {
                    return null;
                }
            }
        }
        #endregion
        #region Methods - Public
        /// <summary>
        /// Adds an IniEntry to an IniSection. If the specified IniSection does not
        /// exist in the current File, it will be added.
        /// </summary>
        /// <param name="section">The IniSection to add to.</param>
        /// <param name="entry">The IniEntry to be added.</param>
        public void AddEntry(IniSection section, IniEntry entry)
        {
            if (this.ContainsSection(section) == false)
            {
                this.AddSection(section);
            }
            section.AddEntry(entry);
        }
        /// <summary>
        /// Adds an IniEntry to an IniSection. If the specified IniSection does not
        /// exist in the current File object's list of IniSection objects,
        /// it will be created.
        /// </summary>
        /// <param name="section">The name of the IniSection to add to.</param>
        /// <param name="entry">The IniEntry to add.</param>
        public void AddEntry(string section, IniEntry entry)
        {
            if (this[section] == null)
            {
                this.CreateSection(section);
            }

            this[section].AddEntry(entry);
        }
        /// <summary>
        /// Adds an IniSection to the current File.
        /// </summary>
        /// <param name="section">The IniSection to add.</param>
        public void AddSection(IniSection section)
        {
            if (this.ContainsSection(section) == false)
            {
                this.sections.Add(section.Name, section);
            }
        }
        /// <summary>
        /// Creates and returns a full clone of the current File,
        /// recursively cloning all IniSections and IniEntries.
        /// </summary>
        /// <returns>A full clone of the current File.</returns>
        public IniFile Clone()
        {
            IniFile clone = new IniFile(this.Location);
            clone.outputSorting = this.outputSorting;
            clone.writeComments = this.writeComments;
            clone.writeEmptySections = this.writeEmptySections;
            clone.unsectioned = this.unsectioned.Clone();

            foreach (IniSection section in this)
            {
                clone.AddSection(section.Clone(clone));
            }

            return clone;
        }
        /// <summary>
        /// Determines whether the specified IniEntry exists in the specified IniSection,
        /// and whether the specified IniSection exists in the current File's IniSection list.
        /// </summary>
        /// <param name="section">The IniSection to check existance for.</param>
        /// <param name="entry">The IniEntry to check existance for.</param>
        /// <returns>True if section exists in the current File's IniSection list,
        /// and entry exists in section's IniEntry list, otherwise returns false.</returns>
        public bool ContainsEntry(IniSection section, IniEntry entry)
        {
            if (this.ContainsSection(section) == true)
            {
                return section.ContainsEntry(entry);
            }
            else
            {
                return false;
            }
        }
        /// <summary>
        /// Determines whether the specified IniEntry exists in the specified IniSection,
        /// and whether the specified IniSection exists in the current File's IniSection list.
        /// </summary>
        /// <param name="section">The IniSection to check existance for.</param>
        /// <param name="entry">The name of the IniEntry to check existance for.</param>
        /// <returns>True if section exists in the current File's IniSection list,
        /// and entry exists in section's IniEntry list, otherwise returns false.</returns>
        public bool ContainsEntry(IniSection section, string entry)
        {
            return section.ContainsEntry(entry);
        }
        /// <summary>
        /// Determines whether the specified IniEntry exists in the specified IniSection,
        /// and whether the specified IniSection exists in the current File's IniSection list.
        /// </summary>
        /// <param name="section">The name of the IniSection to check existance for.</param>
        /// <param name="entry">The IniEntry to check existance for.</param>
        /// <returns>True if section exists in the current File's IniSection list,
        /// and entry exists in section's IniEntry list, otherwise returns false.</returns>
        public bool ContainsEntry(string section, IniEntry entry)
        {
            if (this.ContainsSection(section) == false)
            {
                return false;
            }
            else
            {
                return this[section].ContainsEntry(entry);
            }
        }
        /// <summary>
        /// Determines whether the specified IniEntry exists in the specified IniSection,
        /// and whether the specified IniSection exists in the current File's IniSection list.
        /// </summary>
        /// <param name="section">The name of the IniSection to check existance for.</param>
        /// <param name="entry">The name of the IniEntry to check existance for.</param>
        /// <returns>True if section exists in the current File's IniSection list,
        /// and entry exists in section's IniEntry list, otherwise returns false.</returns>
        public bool ContainsEntry(string section, string entry)
        {
            if (this.ContainsSection(section) == false)
            {
                return false;
            }
            else
            {
                return this[section].ContainsEntry(entry);
            }
        }
        /// <summary>
        /// Determines whether the specified IniSection exists
        /// in the current File's IniSection list.
        /// </summary>
        /// <param name="section">The IniSection to check existance for.</param>
        /// <returns>True if section exists in the current File's IniSection list,
        /// otherwise returns false.</returns>
        public bool ContainsSection(IniSection section)
        {
            if (this[section.Name] == null)
            {
                return false;
            }
            else
            {
                return true;
            }
        }
        /// <summary>
        /// Determines whether the specified IniSection exists
        /// in the current File's IniSection list.
        /// </summary>
        /// <param name="section">The name of the IniSection to check existance for.</param>
        /// <returns>True if section exists in the current File's IniSection list,
        /// otherwise returns false.</returns>
        public bool ContainsSection(string section)
        {
            if (this[section] == null)
            {
                return false;
            }
            else
            {
                return true;
            }
        }
        /// <summary>
        /// Creates an IniEntry in the specified IniSection.
        /// If the specified IniSection does not exist in the current File's IniSection list,
        /// it will be added.
        /// </summary>
        /// <param name="section">The IniSection to create the new IniEntry in.</param>
        /// <param name="entry">The name of the new IniEntry to create.</param>
        public void CreateEntry(IniSection section, string entry)
        {
            if (this.ContainsSection(section) == false)
            {
                this.AddSection(section);
            }
            section.CreateEntry(entry);
        }
        /// <summary>
        /// Creates an IniEntry in the specified IniSection.
        /// If the specified IniSection does not exist in the current File's IniSection list,
        /// it will be created.
        /// </summary>
        /// <param name="section">The name of the IniSection to create the new IniEntry in.</param>
        /// <param name="entry">The name of the new IniEntry to create.</param>
        public void CreateEntry(string section, string entry)
        {
            if (this.ContainsSection(section) == false)
            {
                this.CreateSection(section);
            }
            this[section].CreateEntry(entry);
        }
        /// <summary>
        /// Creates a new IniSection in the current File's IniSection list.
        /// </summary>
        /// <param name="section">The name of the new IniSection to create.</param>
        public IniSection CreateSection(string section)
        {
            if (this.ContainsSection(section) == false)
            {
                IniSection newSection = new IniSection(section, this);
                this.AddSection(newSection);
                return newSection;
            }
            else
            {
                return this[section];
            }
        }
        /// <summary>
        /// Gets an enumerator for the current IniFile.
        /// </summary>
        /// <returns></returns>
        public IEnumerator<IniSection> GetEnumerator()
        {
            foreach (IniSection section in this.sections.Values)
            {
                yield return section;
            }
        }
        /// <summary>
        /// Determines whether the specified string can be used as a valid name
        /// for an IniSection or IniEntry object.
        /// </summary>
        /// <param name="name">The name to validate.</param>
        /// <returns>False if the name paramater contains any invalid chars.</returns>
        public static bool IsValidName(string name)
        {
            if (name.IndexOfAny(IniFile.InvalidNameChars) == -1)
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        /// <summary>
        /// Determines whether the specified name can be used as a valid value
        /// for an IniEntry.
        /// </summary>
        /// <param name="value">The value to validate.</param>
        /// <returns>False if the value paramater contains any invalid chars.</returns>
        public static bool IsValidValue(string value)
        {
            if (value.IndexOfAny(IniFile.InvalidValueChars) == -1)
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        /// <summary>
        /// Parses the INI format file located at the
        /// location property of the current File,
        /// and populates the current File with the parsed data.
        /// </summary>
        public void Read()
        {
            this.sections = new Dictionary<string, IniSection>();
            this.unsectioned = null;

            StreamReader Reader = File.OpenText(this.location);
            IniSection ActiveSection = null;
            IniEntry ActiveEntry = null;
            List<string> commentLines = new List<string>(1);

            int CurrentLine = 0;

            while (true)
            {
                if (Reader.EndOfStream == true)
                {
                    break;
                }
                string line = Reader.ReadLine().Trim();

                if (line == null)
                {
                    break;
                }
                else if (line == String.Empty)
                {
                    continue;
                }
                else if ((line[0] == ';') || (line[0] == '#'))
                {
                    commentLines.Add(line.Substring(1, line.Length - 1).Trim());
                }
                else
                {
                    if (line[0] == '[')
                    {
                        if ((line[line.Length - 1] != ']') || (line.Contains("]") == false))
                        {
                            throw new IniReadException("Invalid section declaration syntax encountered.", CurrentLine, line);
                        }
                        else
                        {
                            if (ActiveSection != null)
                            {
                                this.AddSection(ActiveSection);
                            }
                            
                            string Name = line.Substring(1, line.LastIndexOf(']') - 1);
                            if (IniFile.IsValidName(Name) == false)
                            {
                                throw new IniReadException("Invalid section name.", CurrentLine, line);
                            }
                            else
                            {
                                ActiveSection = new IniSection(Name, this);
                            }
                            if (commentLines.Count > 0)
                            {
                                ActiveSection.Comment = string.Join(Environment.NewLine, commentLines.ToArray());
                                commentLines.Clear();
                                commentLines.Capacity = 1;
                            }
                        }
                    }
                    else
                    {
                        if (line.Contains("=") == false)
                        {
                            throw new IniReadException("Line syntax error.", CurrentLine, line);
                        }
                        else
                        {
                            string[] Parts = line.Split('=');
                            if (IniFile.IsValidName(Parts[0]) == false)
                            {
                                throw new IniReadException("Invalid entry name.", CurrentLine, line);
                            }
                            else
                            {
                                ActiveEntry = new IniEntry(Parts[0].Trim(), this);
                                ActiveEntry.Value = Parts[1].Trim();
                                if (commentLines.Count > 0)
                                {
                                    ActiveEntry.Comment = string.Join(Environment.NewLine, commentLines.ToArray());
                                    commentLines.Clear();
                                    commentLines.Capacity = 1;
                                }
                                

                            }
                            if (ActiveEntry != null)
                            {
                                if (ActiveSection != null)
                                {
                                    this.AddEntry(ActiveSection, ActiveEntry);
                                    //Console.WriteLine("Adding entry to {1} group: {0}", ActiveEntry.ToString(), ActiveSection.ToString());
                                }
                                else
                                {
                                    this.AddEntry(this.NoSection, ActiveEntry);
                                    //Console.WriteLine("Adding entry to default group: {0}", ActiveEntry.ToString());
                                }
                                ActiveEntry = null;
                            }
                        }
                    }
                }
                CurrentLine++;
            }
            Reader.Close();
            Reader.Dispose();
        }
        /// <summary>
        /// Removes an IniEntry from a specified IniSection.
        /// </summary>
        /// <param name="section">The IniSection to remove the IniEntry from.</param>
        /// <param name="entry">The IniEntry to remove.</param>
        public void RemoveEntry(IniSection section, IniEntry entry)
        {
            section.RemoveEntry(entry);
        }
        /// <summary>
        /// Removes an IniEntry from a specified IniSection.
        /// </summary>
        /// <param name="section">The IniSection to remove the IniEntry from.</param>
        /// <param name="entry">The name of the IniEntry to remove.</param>
        public void RemoveEntry(IniSection section, string entry)
        {
            section.RemoveEntry(entry);
        }
        /// <summary>
        /// Removes an IniEntry from a specified IniSection.
        /// </summary>
        /// <param name="section">The name of the IniSection to remove the IniEntry from.</param>
        /// <param name="entry">The IniEntry to remove.</param>
        public void RemoveEntry(string section, IniEntry entry)
        {
            if (this.ContainsSection(section) == true)
            {
                this[section].RemoveEntry(entry);
            }
        }
        /// <summary>
        /// Removes an IniEntry from a specified IniSection.
        /// </summary>
        /// <param name="section">The name of the IniSection to remove the IniEntry from.</param>
        /// <param name="entry">The name of the IniEntry to remove.</param>
        public void RemoveEntry(string section, string entry)
        {
            if (this.ContainsSection(section) == true)
            {
                this[section].RemoveEntry(entry);
            }
        }
        /// <summary>
        /// Removes an IniSection from the current File.
        /// </summary>
        /// <param name="section">The IniSection to remove.</param>
        public void RemoveSection(IniSection section)
        {
            if (this.sections.ContainsKey(section.Name) == true)
            {
                this.sections.Remove(section.Name);
            }
        }
        /// <summary>
        /// Removes an IniSection from the current File.
        /// </summary>
        /// <param name="section">The name of the IniSection to remove.</param>
        public void RemoveSection(string section)
        {
            if (this.ContainsSection(section) == true)
            {
                this.RemoveSection(this[section]);
            }
        }
        /// <summary>
        /// Sets the encoding used for both reading and writing.
        /// </summary>
        /// <param name="encoding"></param>
        public void SetEncoding(Encoding encoding)
        {
            this.readEncoding = encoding;
            this.writeEncoding = encoding;
        }
        /// <summary>
        /// Renames an IniSection.
        /// </summary>
        /// <remarks>
        /// The IniSection being renamed must exist in the current IniFile.
        /// </remarks>
        /// <param name="oldName">The name of the IniSection to rename.</param>
        /// <param name="newName">The new name of the IniSection.</param>
        public void RenameSection(string oldName, string newName)
        {
            if (this.caseSensitiveRetrieval == true)
            {
                oldName = oldName.ToLower();
            }
            if (this.sections.ContainsKey(oldName) == true)
            {
                IniSection section = this.sections[oldName];
                this.sections.Remove(oldName);
                section.Name = newName;
                this.sections.Add((this.caseSensitiveRetrieval == true) ? newName.ToLower() : newName, section);
            }
        }
        /// <summary>
        /// Returns a string representation of the current File.
        /// </summary>
        /// <returns>A string representation of the current File.</returns>
        public override string ToString()
        {
            StringBuilder output = new StringBuilder();

            IEnumerable<IniSection> processedSections;

            if (this.OutputSorting == true)
            {
                System.Collections.Generic.SortedDictionary<string, IniSection> sorted = new SortedDictionary<string, IniSection>();

                foreach (IniSection section in this)
                {
                    sorted.Add(section.Name, section);
                }
                IniSection[] temp = new IniSection[sorted.Count];
                sorted.Values.CopyTo(temp, 0);
                processedSections = temp;
            }
            else
            {
                processedSections = this;
            }
            if (this.unsectioned != null)
            {
                output.Append(this.unsectioned.ToString(false));
            }
            foreach (IniSection Section in processedSections)
            {
                if (this.writeEmptySections == true)
                {
                    output.Append((Section.EntryCount > 0) ? Section.ToString() : "");
                }
                else
                {
                    output.Append(Section.ToString());
                }
            }
            return output.ToString();
        }
        /// <summary>
        /// Writes the INI-format representation of the current File to the file
        /// located at the location property.
        /// </summary>
        public void Write()
        {
            StreamWriter Writer = new StreamWriter(this.location, false, this.writeEncoding);
            Writer.NewLine = Environment.NewLine;

            Writer.Write(this.ToString());

            Writer.Flush();
            Writer.Close();
            Writer.Dispose();
        }
        #endregion
        #region Properties - Public
        /// <summary>
        /// Gets or sets the character used to start comments.
        /// </summary>
        public char CommentChar
        {
            get
            {
                return this.commentChar;
            }
            set
            {
                this.commentChar = value;
            }
        }
        /// <summary>
        /// Gets or sets a value that determines if section and entry names are case-sensitive
        /// for the purpose of retrieving them.
        /// </summary>
        public bool CaseSensitiveRetrieval
        {
            get
            {
                return this.caseSensitiveRetrieval;
            }
            set
            {
                this.caseSensitiveRetrieval = value;
            }
        }
        /// <summary>
        /// Gets the total number of IniEntries contained
        /// in all IniSections in the current File.
        /// </summary>
        public int EntryCount
        {
            get
            {
                int I = 0;
                foreach (IniSection Section in this)
                {
                    I += Section.EntryCount;
                }
                return I;
            }
        }
        /// <summary>
        /// Gets a boolean value that determines whether or not the current File
        /// has a physical representation in the form of a file on the local filesystem.
        /// </summary>
        /// <remarks>If the Write method has been called,
        /// this property should be true, unless the Write method failed.
        /// </remarks>
        public bool ExistsOnDisk
        {
            get
            {
                if (File.Exists(this.location) == false)
                {
                    return false;
                }
                else
                {
                    return true;
                }
            }
        }
        /// <summary>
        /// Gets or sets the location string where the current File should Read from or Write to.
        /// </summary>
        public string Location
        {
            get
            {
                return this.location;
            }
            set
            {
                this.location = value;
            }
        }
        /// <summary>
        /// Gets a pseudo-IniSection that is used to hold any IniEntry objects
        /// that are not actually placed in an IniSection.
        /// </summary>
        public IniSection NoSection
        {
            get
            {
                if (this.unsectioned == null)
                {
                    this.unsectioned = new IniSection("NoSection", this);
                }
                return this.unsectioned;
            }
        }
        /// <summary>
        /// Gets or sets a value that determines whether IniSections will be sorted 
        /// alphabetically when being written to a file or string.
        /// </summary>
        public bool OutputSorting
        {
            get
            {
                return this.outputSorting;
            }
            set
            {
                this.outputSorting = value;
            }
        }
        /// <summary>
        /// Gets or sets the encoding used for reading only.
        /// </summary>
        public Encoding ReadEncoding
        {
            get
            {
                return this.readEncoding;
            }
            set
            {
                this.readEncoding = value;
            }
        }
        /// <summary>
        /// Gets the total number of IniSections contained in the current File.
        /// </summary>
        public int SectionCount
        {
            get
            {
                return this.sections.Count;
            }
        }
        /// <summary>
        /// Gets or sets a boolean value that determines whether or not
        /// comment lines will be included when writing the current File
        /// to a file or string.
        /// </summary>
        /// <remarks>
        /// Setting this value replicates the new value to all child IniSections and IniEntries.
        /// </remarks>
        public bool WriteComments
        {
            get
            {
                return this.writeComments;
            }
            set
            {
                this.writeComments = value;
            }
        }
        /// <summary>
        /// Gets or sets a boolean value that indicates whether or not
        /// empty IniSection objects will be included when writing the
        /// current File to a file or string.
        /// </summary>
        public bool WriteEmptySections
        {
            get
            {
                return this.writeEmptySections;
            }
            set
            {
                this.writeEmptySections = value;
            }
        }
        /// <summary>
        /// Gets or sets the encoding used for writing only.
        /// </summary>
        public Encoding WriteEncoding
        {
            get
            {
                return this.writeEncoding;
            }
            set
            {
                this.writeEncoding = value;
            }
        }
        #endregion
        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        #endregion
    }
}
