/*
LibINI - INI-format I/O library for the CLR.
Copyright  2006-2007 Serenity Project  

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace LibINI
{
    /// <summary>
    /// Represents a grouping of key/value pairs (IniEntry objects).
    /// </summary>
    public sealed class IniSection : IEnumerable<IniEntry>
    {
        #region Constructors - Internal
        internal IniSection(string name, IniFile file)
        {
            this.name = name;
            this.file = file;
            this.entries = new Dictionary<string, IniEntry>();
        }
        #endregion
        #region Fields - Private
        private Dictionary<string, IniEntry> entries;
        private IniEntry lastEntry;
        private string comment;
        private string name;
        private IniFile file;
        #endregion
        #region Indexers - Public
        /// <summary>
        /// Gets an IniEntry contained in the current IniSection.
        /// </summary>
        /// <param name="name">The name of the IniEntry to get.</param>
        /// <returns>An IniEntry named name.</returns>
        public IniEntry this[string name]
        {
            get
            {
                name = name.ToLower();
                if ((this.lastEntry != null) && (name == this.lastEntry.Name.ToLower()))
                {
                    return this.lastEntry;
                }
                else if (this.entries.ContainsKey(name) == true)
                {
                    this.lastEntry = this.entries[name];
                    return this.lastEntry;
                }
                else
                {
                    return null;
                }
            }
        }
        #endregion
        #region Methods - Internal
        internal string ToString(bool includeSection)
        {
            if ((this.file.WriteEmptySections == false) && (this.EntryCount == 0))
            {
                return "";
            }
            else
            {
                StringBuilder output = new StringBuilder();

                if (this.file.WriteComments == true)
                {
                    output.Append(this.comment.Replace(Environment.NewLine, Environment.NewLine + this.file.CommentChar.ToString()));
                }

                if (includeSection == true)
                {
                    output.Append("[" + this.name + "]");
                }
                IEnumerable<IniEntry> processedEntries;
                if (this.file.OutputSorting == true)
                {
                    SortedDictionary<string, IniEntry> sorted = new SortedDictionary<string, IniEntry>(this.entries);

                    processedEntries = sorted.Values;
                }
                else
                {
                    processedEntries = this;
                }
                foreach (IniEntry entry in processedEntries)
                {
                    output.Append(entry.ToString());
                }

                return output.ToString();
            }
        }

        #endregion
        #region Methods - Public
        /// <summary>
        /// Adds an IniEntry object to the current IniSection.
        /// </summary>
        /// <param name="entry">The IniEntry object to be added.</param>
        public void AddEntry(IniEntry entry)
        {
            string name = entry.Name.ToLower();
            if (this.entries.ContainsKey(name) == false)
            {
                this.entries.Add(name, entry);
            }
        }
        /// <summary>
        /// Creates a deep copy of the current IniSection.
        /// </summary>
        /// <returns></returns>
        public IniSection Clone()
        {
            return this.Clone(this.File);
        }
        /// <summary>
        /// Creates a deep copy of the current IniSection.
        /// </summary>
        /// <returns>A full clone of the current IniSection.</returns>
        public IniSection Clone(IniFile file)
        {
            IniSection clone = new IniSection(this.Name, file);
            clone.comment = this.comment;
            foreach (IniEntry entry in this)
            {
                clone.entries.Add(entry.Name.ToLower(), entry);
            }
            return clone;
        }
        /// <summary>
        /// Creates a new IniEntry using the specified name,
        /// adds it to the current IniSection, and returns it for further use.
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public IniEntry CreateEntry(string name)
        {
            if (this.file.CaseSensitiveRetrieval == true)
            {
                name = name.ToLower();
            }
            IniEntry newEntry = new IniEntry(name, this.File);
            this.entries.Add(name.ToLower(), newEntry);
            return newEntry;
        }
        /// <summary>
        /// Creates a new IniEntry using the specified name and value,
        /// adds it to the current IniSection, and returns it for further use.
        /// </summary>
        /// <param name="name">The name of the new IniEntry.</param>
        /// <param name="value">A value to initialize the new IniEntry with.</param>
        public IniEntry CreateEntry(string name, string value)
        {
            IniEntry newEntry = new IniEntry(name, this.File);
            newEntry.Value = value;
            this.entries.Add(name.ToLower(), newEntry);
            return newEntry;
        }
        /// <summary>
        /// Determines whether an IniEntry exists in the current IniSection.
        /// </summary>
        /// <param name="entry">The IniEntry object to check existance for.</param>
        /// <returns>True if the object is in the current IniSection's IniEntry list, otherwise false.</returns>
        public bool ContainsEntry(IniEntry entry)
        {
            string name;
            if (this.file.CaseSensitiveRetrieval == true)
            {
                name = entry.Name.ToLower();
            }
            else
            {
                name = entry.Name;
            }
            return this.entries.ContainsKey(name);
        }
        /// <summary>
        /// Determines whether an IniEntry exists in the current IniSection.
        /// </summary>
        /// <param name="entry">The name of the IniEntry object to check existance for.</param>
        /// <returns>True if the object is in the current IniSection's IniEntry list, otherwise false.</returns>
        public bool ContainsEntry(string entry)
        {
            string name;
            if (this.file.CaseSensitiveRetrieval == true)
            {
                name = entry.ToLower();
            }
            else
            {
                name = entry;
            }
            return this.entries.ContainsKey(name);
        }
        /// <summary>
        /// Gets an enumerator for the current IniSection.
        /// </summary>
        /// <returns></returns>
        public IEnumerator<IniEntry> GetEnumerator()
        {
            foreach (IniEntry entry in this.entries.Values)
            {
                yield return entry;
            }
        }
        /// <summary>
        /// Removes an IniEntry from the current IniSection.
        /// </summary>
        /// <param name="entry">The IniEntry to remove.</param>
        public void RemoveEntry(IniEntry entry)
        {
            string name;
            if (this.file.CaseSensitiveRetrieval == true)
            {
                name = entry.Name.ToLower();
            }
            else
            {
                name = entry.Name;
            }
  
            if (this.entries.ContainsKey(name) == true)
            {
                this.entries.Remove(name);
            }
        }
        /// <summary>
        /// Removes an IniEntry from the current IniSection.
        /// </summary>
        /// <param name="entry">The name of the IniEntry to remove.</param>
        public void RemoveEntry(string entry)
        {
            string name;
            if (this.file.CaseSensitiveRetrieval == true)
            {
                name = entry.ToLower();
            }
            else
            {
                name = entry;
            }
            if (this.ContainsEntry(name) == true)
            {
                this.entries.Remove(name);
            }
        }
        /// <summary>
        /// Renames an IniEntry.
        /// </summary>
        /// <remarks>
        /// The IniEntry must exist in the current IniSection.
        /// </remarks>
        /// <param name="oldName">The name of the IniEntry to rename.</param>
        /// <param name="newName">The new name of the IniEntry.</param>
        public void RenameEntry(string oldName, string newName)
        {
            if (this.file.CaseSensitiveRetrieval == true)
            {
                oldName = oldName.ToLower();
            }
            if (this.entries.ContainsKey(oldName) == true)
            {
                IniEntry entry = this.entries[oldName];
                this.entries.Remove(oldName);
                entry.Name = newName;
                this.entries.Add((this.file.CaseSensitiveRetrieval == true) ? newName.ToLower() : newName, entry);
            }
        }
        /// <summary>
        /// Creates and returns an INI-format file fragment representing the
        /// current IniSection and it's list of IniEntry objects.
        /// </summary>
        /// <returns>A string containing an INI-format file fragment.</returns>
        public override string ToString()
        {
            return this.ToString(true);
        }
        #endregion
        #region Properties - Public
        /// <summary>
        /// Gets or sets a comment associated with the current IniSection.
        /// </summary>
        public string Comment
        {
            get
            {
                return this.comment;
            }
            set
            {
                this.comment = value;
            }
        }
        /// <summary>
        /// Gets the number of IniEntries in the current IniSection.
        /// </summary>
        public int EntryCount
        {
            get
            {
                return this.entries.Count;
            }
        }
        /// <summary>
        /// Gets the IniFile that the current IniSection is a part of.
        /// </summary>
        public IniFile File
        {
            get
            {
                return this.File;
            }
            internal set
            {
                this.file = value;
            }
        }
        /// <summary>
        /// Gets the name of the current IniElement.
        /// </summary>
        public string Name
        {
            get
            {
                return this.name;
            }
            internal set
            {
                if (IniFile.IsValidName(value) == true)
                {
                    this.name = value;
                }
                else
                {
                    throw new IniOperationException("An attempt was made to assign an invalid name to an IniSection.", value);
                }
            }
        }
        #endregion
        #region IEnumerable Members

        /// <summary>
        /// Gets a nongeneric enumerator for the current IniSection.
        /// </summary>
        /// <returns></returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            foreach (IniEntry entry in this.entries.Values)
            {
                yield return entry;
            }
        }

        #endregion
    }
}
